#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import testtools

from tests.functional import utils


class DeploymentTestCase(testtools.TestCase):

    def test_create_deployment_from_env(self):
        os.environ.update(
            {
                "OS_AUTH_URL": "http://fake",
                "OS_USERNAME": "fake",
                "OS_PASSWORD": "fake"
            }
        )
        rally = utils.Rally(plugin_path="tests/functional/extra")
        rally("deployment create --name fromenv --fromenv")
        spec = rally("deployment config", getjson=True)
        self.assertEqual({
            "good@fake": {
                "auth_url": "http://fake",
                "username": "fake",
                "password": "fake"
            }
        }, spec)
