# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Deleted worker table

Revision ID: 44169f4d455e
Revises: dc46687661df
Create Date: 2018-01-08 21:22:41.682881

"""

from alembic import op

from rally import exceptions

# revision identifiers, used by Alembic.
revision = "44169f4d455e"
down_revision = "dc46687661df"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table("workers")


def downgrade():
    raise exceptions.DowngradeNotSupported()
