{% extends "/base.html" %}

{% block title_text %}Rally Verification Job Results{% endblock %}

{% block css %}
    li { margin:2px 0 }
{% if verifications|length == 1 %}
    ol {padding: 0; list-style-type: none;}
{% endif %}
    a, a:visited { color:#039 }
    code { padding:0 15px; color:#888; display: block }
    .columns li { position:relative }
    .columns li > :first-child { display:block }
    .columns li > :nth-child(2) { display:block; position:static; left:165px; top:0; white-space:nowrap }
    .fail {color: #cc0000}
    .error {color: #cc0000}
    .success {color: green}
    .skip {color: #A9A9A9}
{% endblock %}

{% block css_content_wrap %}margin:0 auto; padding:0 5px{% endblock %}

{% block media_queries %}
    @media only screen and (min-width: 320px) { .content-wrap { width:400px } }
    @media only screen and (min-width: 520px) { .content-wrap { width:500px } }
    @media only screen and (min-width: 620px) { .content-wrap { width:90% } .columns li > :nth-child(2) { position:absolute } }
    @media only screen and (min-width: 720px) { .content-wrap { width:70% } }
{% endblock %}

{% block header_text %}Verify Job Results{% endblock %}

{% block content %}
    <h2>Logs and Results Files</h2>
    <ul class="columns">
      <li><a href="console.html">Job logs</a> <code>console.html</code>
      <li><a href="logs/">Logs of all services</a> <code>logs/</code>
      <li><a href="rally-verify/">Results files</a> <code>rally-verify/</code>
    </ul>

    <h2>Steps</h2>

{% for step in steps %}
    {{ step|safe }}
{% endfor %}

    <h2>About Rally</h2>
    <p>Rally is benchmarking and verification system for OpenStack:</p>
    <ul>
      <li><a href="https://opendev.org/openstack/rally">Git
      repository</a> (<a href="https://github.com/openstack/rally">GitHub mirror</a>)
      <li><a href="https://rally.readthedocs.io/en/latest/">Documentation</a>
      <li><a href="https://rally.readthedocs.io/en/latest/quick_start/index.html">Quick start</a>
      <li><a href="https://rally.readthedocs.io/en/latest/quick_start/gates.html">How to add Rally job to your project</a>
    </ul>
{% endblock %}
