# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jsonschema
import mock

from rally.verification import reporter
from tests.unit import test


class ReporterTestCase(test.TestCase):

    def test_make(self):
        reporter_cls = mock.Mock()

        reporter_cls.return_value.generate.return_value = {}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": {}}
        reporter.VerificationReporter.make(reporter_cls, None, None)
        reporter_cls.return_value.generate.return_value = {
            "files": {"/path/foo": "content"}}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"open": "/path/foo"}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"print": "foo"}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {
            "files": {"/path/foo": "content"}, "open": "/path/foo",
            "print": "foo"}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": []}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": ""}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": {"a": {}}}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"open": []}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"print": []}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"additional": ""}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)
